import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from pathlib import Path
from common.lut import forward_2x2_input_SxS_output, forward_unfolded_2x2_input_SxS_output
from common.layers import PercievePattern

class SDYLutx1(nn.Module):
    def __init__(
        self, 
        quantization_interval,
        scale
    ):
        super(SDYLutx1, self).__init__()     
        self.scale = scale
        self.quantization_interval = quantization_interval
        self._extract_pattern_S = PercievePattern(receptive_field_idxes=[[0,0],[0,1],[1,0],[1,1]], center=[0,0], window_size=3)
        self._extract_pattern_D = PercievePattern(receptive_field_idxes=[[0,0],[2,0],[0,2],[2,2]], center=[0,0], window_size=3)
        self._extract_pattern_Y = PercievePattern(receptive_field_idxes=[[0,0],[1,1],[1,2],[2,1]], center=[0,0], window_size=3)
        self.stageS = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))
        self.stageD = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))
        self.stageY = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))

    @staticmethod
    def init_from_lut(
        stageS, stageD, stageY
    ):   
        scale = int(stageS.shape[-1])
        quantization_interval = 256//(stageS.shape[0]-1)
        lut_model = SDYLutx1(quantization_interval=quantization_interval, scale=scale)
        lut_model.stageS = nn.Parameter(torch.tensor(stageS).type(torch.float32))
        lut_model.stageD = nn.Parameter(torch.tensor(stageD).type(torch.float32))
        lut_model.stageY = nn.Parameter(torch.tensor(stageY).type(torch.float32))
        return lut_model

    def forward(self, x):
        b,c,h,w = x.shape
        x = x.view(b*c, 1, h, w).type(torch.float32)
        output = torch.zeros([b*c, 1, h*self.scale, w*self.scale], dtype=x.dtype, device=x.device)
        for rotations_count in range(4):
            rotated = torch.rot90(x, k=rotations_count, dims=[-2, -1])
            rb,rc,rh,rw = rotated.shape

            s = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_S(rotated), lut=self.stageS)         
            s = s.view(rb*rc, 1, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb*rc, 1, rh*self.scale, rw*self.scale) 
            s = torch.rot90(s, k=-rotations_count, dims=[-2, -1])            
            output += s

            d = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_D(rotated), lut=self.stageD)            
            d = d.view(rb*rc, 1, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb*rc, 1, rh*self.scale, rw*self.scale) 
            d = torch.rot90(d, k=-rotations_count, dims=[-2, -1])
            output += d

            y = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_Y(rotated), lut=self.stageY)            
            y = y.view(rb*rc, 1, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb*rc, 1, rh*self.scale, rw*self.scale) 
            y = torch.rot90(y, k=-rotations_count, dims=[-2, -1])
            output += y
                       
        output /= 4*3
        output = output.view(b, c, h*self.scale, w*self.scale)
        return output

    def __repr__(self):
        return f"{self.__class__.__name__}" + \
               f"\n  stageS size: {self.stageS.shape}" + \
               f"\n  stageD size: {self.stageD.shape}" + \
               f"\n  stageY size: {self.stageY.shape}"



class SDYLutx2(nn.Module):
    def __init__(
        self, 
        quantization_interval,
        scale
    ):
        super(SDYLutx2, self).__init__()     
        self.scale = scale
        self.quantization_interval = quantization_interval
        self._extract_pattern_S = PercievePattern(receptive_field_idxes=[[0,0],[0,1],[1,0],[1,1]], center=[0,0], window_size=2)
        self._extract_pattern_D = PercievePattern(receptive_field_idxes=[[0,0],[2,0],[0,2],[2,2]], center=[0,0], window_size=3)
        self._extract_pattern_Y = PercievePattern(receptive_field_idxes=[[0,0],[1,1],[1,2],[2,1]], center=[0,0], window_size=3)
        self.stage1_S = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (1,1)).type(torch.float32))
        self.stage1_D = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (1,1)).type(torch.float32))
        self.stage1_Y = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (1,1)).type(torch.float32))
        self.stage2_S = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))
        self.stage2_D = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))
        self.stage2_Y = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))

    @staticmethod
    def init_from_lut(
        stage1_S, stage1_D, stage1_Y, stage2_S, stage2_D, stage2_Y
    ):   
        scale = int(stage2_S.shape[-1])
        quantization_interval = 256//(stage2_S.shape[0]-1)
        lut_model = SDYLutx2(quantization_interval=quantization_interval, scale=scale)
        lut_model.stage1_S = nn.Parameter(torch.tensor(stage1_S).type(torch.float32))
        lut_model.stage1_D = nn.Parameter(torch.tensor(stage1_D).type(torch.float32))
        lut_model.stage1_Y = nn.Parameter(torch.tensor(stage1_Y).type(torch.float32))
        lut_model.stage2_S = nn.Parameter(torch.tensor(stage2_S).type(torch.float32))
        lut_model.stage2_D = nn.Parameter(torch.tensor(stage2_D).type(torch.float32))
        lut_model.stage2_Y = nn.Parameter(torch.tensor(stage2_Y).type(torch.float32))
        return lut_model

    def forward(self, x):
        b,c,h,w = x.shape
        x = x.view(b*c, 1, h, w).type(torch.float32)
        output = torch.zeros([b*c, 1, h, w], dtype=x.dtype, device=x.device) 
        for rotations_count in range(4):
            rotated = torch.rot90(x, k=rotations_count, dims=[-2, -1])
            rb,rc,rh,rw = rotated.shape

            s = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_S(rotated), lut=self.stage1_S)         
            s = s.view(rb, rc, rh, rw, 1, 1).permute(0,1,2,4,3,5).reshape(rb, rc, rh, rw) 
            s = torch.rot90(s, k=-rotations_count, dims=[-2, -1])            
            output += s

            d = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_D(rotated), lut=self.stage1_D)            
            d = d.view(rb, rc, rh, rw, 1, 1).permute(0,1,2,4,3,5).reshape(rb, rc, rh, rw) 
            d = torch.rot90(d, k=-rotations_count, dims=[-2, -1])
            output += d

            y = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_Y(rotated), lut=self.stage1_Y)            
            y = y.view(rb, rc, rh, rw, 1, 1).permute(0,1,2,4,3,5).reshape(rb, rc, rh, rw) 
            y = torch.rot90(y, k=-rotations_count, dims=[-2, -1])
            output += y
                       
        output /= 4*3
        x = output

        output = torch.zeros([b*c, 1, h*self.scale, w*self.scale], dtype=x.dtype, device=x.device)
        for rotations_count in range(4):
            rotated = torch.rot90(x, k=rotations_count, dims=[-2, -1])
            rb,rc,rh,rw = rotated.shape

            s = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_S(rotated), lut=self.stage2_S)         
            s = s.view(rb, rc, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb, rc, rh*self.scale, rw*self.scale) 
            s = torch.rot90(s, k=-rotations_count, dims=[-2, -1])            
            output += s

            d = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_D(rotated), lut=self.stage2_D)            
            d = d.view(rb, rc, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb, rc, rh*self.scale, rw*self.scale) 
            d = torch.rot90(d, k=-rotations_count, dims=[-2, -1])
            output += d

            y = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_Y(rotated), lut=self.stage2_Y)            
            y = y.view(rb, rc, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb, rc, rh*self.scale, rw*self.scale) 
            y = torch.rot90(y, k=-rotations_count, dims=[-2, -1])
            output += y
                       
        output /= 4*3
        output = output.view(b, c, h*self.scale, w*self.scale)
        
        return output

    def __repr__(self):
        return f"{self.__class__.__name__}" + \
               f"\n  stage1_S size: {self.stage1_S.shape}" + \
               f"\n  stage1_D size: {self.stage1_D.shape}" + \
               f"\n  stage1_Y size: {self.stage1_Y.shape}" + \
               f"\n  stage2_S size: {self.stage2_S.shape}" + \
               f"\n  stage2_D size: {self.stage2_D.shape}" + \
               f"\n  stage2_Y size: {self.stage2_Y.shape}"



class SDYLutCenteredx1(nn.Module):
    def __init__(
        self, 
        quantization_interval,
        scale
    ):
        super(SDYLutCenteredx1, self).__init__()     
        self.scale = scale
        self.quantization_interval = quantization_interval
        self._extract_pattern_S = PercievePattern(receptive_field_idxes=[[0,0],[0,1],[1,0],[1,1]], center=[1,1], window_size=3)
        self._extract_pattern_D = PercievePattern(receptive_field_idxes=[[0,0],[2,0],[0,2],[2,2]], center=[1,1], window_size=3)
        self._extract_pattern_Y = PercievePattern(receptive_field_idxes=[[0,0],[1,1],[1,2],[2,1]], center=[1,1], window_size=3)
        self.stageS = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))
        self.stageD = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))
        self.stageY = nn.Parameter(torch.randint(0, 255, size=(256//quantization_interval+1,)*4 + (scale,scale)).type(torch.float32))

    @staticmethod
    def init_from_lut(
        stageS, stageD, stageY
    ):   
        scale = int(stageS.shape[-1])
        quantization_interval = 256//(stageS.shape[0]-1)
        lut_model = SDYLutCenteredx1(quantization_interval=quantization_interval, scale=scale)
        lut_model.stageS = nn.Parameter(torch.tensor(stageS).type(torch.float32))
        lut_model.stageD = nn.Parameter(torch.tensor(stageD).type(torch.float32))
        lut_model.stageY = nn.Parameter(torch.tensor(stageY).type(torch.float32))
        return lut_model

    def forward(self, x):
        b,c,h,w = x.shape
        x = x.view(b*c, 1, h, w).type(torch.float32)
        output = torch.zeros([b*c, 1, h*self.scale, w*self.scale], dtype=x.dtype, device=x.device)
        for rotations_count in range(4):
            rotated = torch.rot90(x, k=rotations_count, dims=[-2, -1])
            rb,rc,rh,rw = rotated.shape

            s = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_S(rotated), lut=self.stageS)         
            s = s.view(rb*rc, 1, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb*rc, 1, rh*self.scale, rw*self.scale) 
            s = torch.rot90(s, k=-rotations_count, dims=[-2, -1])            
            output += s

            d = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_D(rotated), lut=self.stageD)            
            d = d.view(rb*rc, 1, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb*rc, 1, rh*self.scale, rw*self.scale) 
            d = torch.rot90(d, k=-rotations_count, dims=[-2, -1])
            output += d

            y = forward_unfolded_2x2_input_SxS_output(index=self._extract_pattern_Y(rotated), lut=self.stageY)            
            y = y.view(rb*rc, 1, rh, rw, self.scale, self.scale).permute(0,1,2,4,3,5).reshape(rb*rc, 1, rh*self.scale, rw*self.scale) 
            y = torch.rot90(y, k=-rotations_count, dims=[-2, -1])
            output += y
                       
        output /= 4*3
        output = output.view(b, c, h*self.scale, w*self.scale)
        return output

    def __repr__(self):
        return f"{self.__class__.__name__}" + \
               f"\n  stageS size: {self.stageS.shape}" + \
               f"\n  stageD size: {self.stageD.shape}" + \
               f"\n  stageY size: {self.stageY.shape}"